from .api import APIClient
from .client import DockerClient, from_context, from_env
from .context import Context, ContextAPI
from .tls import TLSConfig
from .version import __version__

__title__ = 'docker'
