dl.hide-signature > dt {
  display: none;
}

dl.field-list > dt {
    /* prevent code blocks from forcing wrapping on the "Parameters" header */
    word-break: initial;
}

code.literal{
  hyphens: none;
}
